/* RFID Safety Sensor - Hazard Perimter
   Assad Ebrahim, v8c
   2019-06-02--Sun--17:57
   This program sends serial commands to the Arduino NANO microcontroller and ThingMagic M6E-NANO RFID Reader, Rbased on user interaction through a menu.
   ESC to exit the program.   
   Mods:
   + cranked up tag scanning power to 27dBm
   ! NOTE: To further reduce latency to tag detection, remove printing of ...s when scanning.
   ! NOTE: To prevent stalling, 
       1) put 100uF power cap back in to smooth power draw when scanning.
       2) externalize heat sink outside pelican box
       3) provide more stable power regulation instead of just the AC wall socket
*/

#include <stdio.h>    // printf() and friends
#include <conio.h>    // getch()/getche()
#include <windows.h>  // Win32 API  (includes winbase.h has serial port definitions, and kernel32)

#define ESC_KEY 0x1B  // 27d
#define EXCL_KEY 0x21  // 33d  (!)
#define HASH_KEY 0x23  // 35d  (#)
#define LF_KEY  0x0A  // 10d
#define CR_KEY  0x0D  // 13d
#define BKSP_KEY 0x08  // 8d
#define DEL_KEY  0x7F  // 127d

// function prototypes
int setupSerial(HANDLE* hSerial, char* portName, DWORD baud);
int sendSerial(HANDLE hSerial, char* msg, int num_bytes);
int readSerial(HANDLE hSerial, unsigned char* SerialBuffer);
unsigned char showMenu();
void alarmOn();
void alarmOff();
void printHexMsg(unsigned char *msg, int len);
unsigned char scanForTag(unsigned char *msg, int len);
void pause();

// globals
  HANDLE WINAPI hSerialUC;
  unsigned char SerialBuffer[256]; //Circular Buffer for storing received data
//  unsigned char msg_AlarmOn[2] = {0xF7, 0x7F};
//  unsigned char msg_AlarmOff[2] = {0xF1, 0x1F};
//  unsigned char msg_resetRFID[2] = {0x22, 0x2F};
  unsigned char msg_AlarmOnCode[1] = {0x21};
  unsigned char msg_AlarmOffCode[1] = {0x23};
  unsigned char msg_resetRFIDCode[1] = {0x41};
  int flag_scanning=0;
  int flag_tagfound=0;
  
int main()
{  
  // Variables
  char portNameUC[]="COM5";   // microcontroller
  int tot_bytes_read=0;
  unsigned char ch;   // user keypress
  
  unsigned char msg_readerStartCode[1] = {0x39} ;
  unsigned char msg_readerStopCode[1] = {0x30} ;

// Setup RFID Serial links with RFID and UC, returning error condition.
  if (1==setupSerial(&hSerialUC, portNameUC, CBR_115200)) return 1;


  printf("Lite Controller for RFID Tag Reader.  (c) June 2019, Assad Ebrahim\n");
  printf("----------------------------------------------------------------------\n");
  
  ch=showMenu();  // menu choices are numbers 1-9, 0, and !, #
  do
  {
     switch(ch) 
     {
       // manual alarm on/off for testing
       case EXCL_KEY : alarmOn(); break;
       case HASH_KEY : alarmOff(); break;   // there is also a mechanical alarmOff pushbutton wired into the circuitry
       case 0x39 :
          printf("Sending readerStartScanning()... ");    
          flag_scanning = 1;
          sendSerial(hSerialUC, msg_readerStartCode, 1); 
          break;
       case 0x30 :
          printf("Sending readerStopScanning()... ");     
          flag_scanning = 0;
          sendSerial (hSerialUC, msg_readerStopCode, 1);
          break;
       default: break;
       // character not from options in menu.  Continue with previous state / return to reading serial port until next key is pressed
       // pass through to reading serial port
    } // end switch

    // Read serial port until key is pressed
    printf("Reading serial port... \n");
    while(kbhit()==0)   // poll keyboard for key press
    {  
        tot_bytes_read = readSerial(hSerialUC, SerialBuffer);
        if (tot_bytes_read>0) { 
            // pause();
            if(flag_scanning==0) {
                printHexMsg(SerialBuffer, tot_bytes_read); 
            } else { printf("."); }
            scanForTag(SerialBuffer, tot_bytes_read-2); 
        }
    }
    if(ch!=ESC_KEY) 
    {
        printf("\nPaused.  Press key to resume reading, or Menu key...\n");
        ch=showMenu();
    }
  } while (ch!=ESC_KEY) ;  // ESC to exit loop, otherwise show menu and resume
  printf("**Halting at user request. **\n");


  // Close serial port UC
    printf("Closing UC serial ports...");
    if (CloseHandle(hSerialUC) == 0)
    {  printf("Error!\n"); return 1; }
    else printf("OK.\n");
    sleep(200);

  return 0;  // exit normally
}

/* ***************************  Subroutines ************************* */

unsigned char showMenu() 
{
  printf("Menu: 9-startScanning, 0-stopScanning, !-Alarm ON Test, #-Alarm OFF Test, A-Reset RFID\n");
  printf("(Choose number, or press ESC to exit program or any other key to resume scanning the serial port (subsequently any key to see menu again)\n");
  printf("--------------------------------------------------------------------\n");
  return getch();
}

void alarmOn( ) {
    printf("\n**ALARM ON!** (sending signal)\n"); 
//    sendSerial (hSerialUC, msg_AlarmOnCode, 1);
    // pause();
}

void alarmOff()
{
    printf("\n**ALARM OFF** (sending signal)\n"); 
    sendSerial (hSerialUC, msg_AlarmOffCode, 1);
    // pause();
}

void printHexMsg(unsigned char *msg, int len)
{
    int i=0;
    for(i=0; i<len; i++) { printf("%02x ", msg[i]); }
    printf("\n");
    return;
}

void pause()
{
    printf("Press any key to continue!\n"); 
    getch(); // pause
}

int readSerial(HANDLE hSerial, unsigned char* SerialBuffer)
{
    unsigned char TempChar; //Temporary character used for reading
    DWORD NumBytesRead=0;
    int i=0, j=0;  // every call to Serial starts from the beginning of the circular buffer - this is CRITICAL so that we know where to begin reading from
 
       do
      {
        if (!ReadFile( hSerial,         //Handle of the Serial port
                    &TempChar,       //Temporary character
                    sizeof(TempChar),//Size of TempChar
                    &NumBytesRead,    //Number of bytes read
                    NULL) ) 
        {
            printf("Error!\n");
            printf("Closing serial port...");
            if (CloseHandle(hSerial) == 0)
            {  printf("Error!\n"); return 1; }
            else printf("OK.\n");
            sleep(200);
        }      
        // else printf("Reading OK (%i bytes read).\n", NumBytesRead); 
        if (NumBytesRead > 0) 
        {
          SerialBuffer[i] = TempChar;// Store Tempchar into buffer
          i++;
          j++;  // total number of bytes read
          i=i%256;  // make it a circular buffer --- will need to read out the buffer quickly and handle it
          // putchar(TempChar);  // ASCII character
          // printf(" %02x ", TempChar); // Hex value, 1 byte
          // printf(" (%i,%i) ",i,j); // buffer position
        }
      } while (NumBytesRead > 0);

      if ( j > 0 && flag_scanning == 0) printf("%i bytes read (%i in 256 bytes circular buffer)\n",j,i);  // v7 AE
      return j;
}


unsigned char scanForTag(unsigned char *msg, int len)
{
    int k=0; // index var 0 -- always start reading from beginning of circular buffer
    for (k=0; k<len; k++) {
                // printf(" %i: ", k);   // v7 AE
                if (SerialBuffer[k] == 0xFF && SerialBuffer[k+1] > 0x14 && SerialBuffer[k+2] == 0x22) { // it's a tag!!!
                    // pause();
                    alarmOn();
                    printf(" Tag at [%i]: %02x %02x %02x \n", k, SerialBuffer[k], SerialBuffer[k+1], SerialBuffer[k+2]);
                    printf(" Tag: ");
                    printHexMsg(SerialBuffer+k, 7+SerialBuffer[k+1]);  // print tag
                    // Act on Tag -- insert code here or in main body
                    return (7+SerialBuffer[k+1]); // stop after finding first tag, return next position in buffer
                }
    }
}


int sendSerial(HANDLE hSerial, char* msg, int num_bytes)
{ 
    // sends mesage to specified serial handle (assuming it's been opened), confirms bytes actually sent, and shows message on screen
    DWORD bytes_written=0;
    if(!WriteFile(hSerial, msg, num_bytes, &bytes_written, NULL)) { printf("Error\n"); CloseHandle(hSerial); return 1; }   
    printf("%d bytes written\n", bytes_written);
    printHexMsg(msg, num_bytes);
    sleep(200);  // delay to ensure message is received
    return 0;
}


int setupSerial(HANDLE* hSerial, char* portName, DWORD baud)
{
  // Serial variables
  DCB dcbSerialParams = {0};
  COMMTIMEOUTS timeouts = {0};
        
  // Setting up Serial Port  
  printf("Opening serial port: %s ... ", portName);
  *hSerial = CreateFile( portName, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
  if (*hSerial==INVALID_HANDLE_VALUE) { printf("Error!\n"); } else printf("OK.\n");
  sleep(200);
    
  //Getting device parameters
  printf("Getting device state... ");
  dcbSerialParams.DCBlength = sizeof(dcbSerialParams);
  if (GetCommState(*hSerial, &dcbSerialParams) == 0)
    {
        fprintf(stderr, "Error!\n");
        CloseHandle(*hSerial);
        return 1; // error
    } else printf("OK.\n");
  sleep(200);
  
  //Setting device parameters
    dcbSerialParams.BaudRate = baud; //CBR_9600;
    dcbSerialParams.ByteSize = 8;
    dcbSerialParams.StopBits = ONESTOPBIT;
    dcbSerialParams.Parity = NOPARITY;
  
  // Configure port
  printf("Configuring port (setting device parameters)... ");
 if(SetCommState(*hSerial, &dcbSerialParams) == 0)
    {
        printf("Error!\n");
        CloseHandle(*hSerial);
        return 1;
    } else printf("OK.\n");
  sleep(200);
  
  // Setting COM port timeout settings
  printf("Setting timeouts... ");
    timeouts.ReadIntervalTimeout = 50;
    timeouts.ReadTotalTimeoutConstant = 50;
    timeouts.ReadTotalTimeoutMultiplier = 10;
    timeouts.WriteTotalTimeoutConstant = 50;
    timeouts.WriteTotalTimeoutMultiplier = 10;
  if(SetCommTimeouts(*hSerial, &timeouts) == 0)
    {
        printf("Error!\n");
        CloseHandle(*hSerial);
        return 1;
    } else printf("OK.\n");
  sleep(200);

    return 0;  // return normally
}





/* TO DO LIST
- Alarm On: turn on buzzer, then later externally powered siren
- Alarm Off: what process?  push button? (signal)?
- Recognize a tag
- Submit tag for exclusion (driver)
- Test process for tag recognitions
- Distance / signal strength
- Tag ID - which tag?

- battery
- Antenna
- tags
- relay?
- design out the laptop (Rx/Tx comm with less noise/error in serial comms between RFID and uC)

- buy batteries Li-ion; 18* cells 4000mAh
- buy flashing light siren

*/